/*
 *      chlist.c
 *      
 *      Copyright 2009 - 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glib/gi18n.h>
#include <string.h>
#include "dbstorage.h"
#include "svcparams.h"
#include "chlist.h"
#include "udpserver.h"
#include "mainwindow.h"
#include "dvbstatus.h"
#include "resources.h"


struct _servicelist {
	GtkWidget *toolbar;
	GtkToolItem *tb_new, *tb_edit, *tb_remove, *tb_netreload, *tb_netclear, *tb_separator;
	GtkWidget *icon_view; GtkListStore *store;
	GtkWidget *pm_menu, *pm_new, *pm_edit, *pm_delete;
	GtkAccelGroup *pm_group;
} servicelist;

static GtkWidget *window = NULL;

static guint32 current_service = 0;
static guchar current_service_type = SERVICE_TRANSPONDER;

enum {
	COL_TEXT,
	COL_PIXBUF,
	COL_ID,
	COL_TYPE,
	NUM_COLS
};

struct _sindex {
	guint32 *vec;
	gsize count;
	guint32 ptr;
};

static guint servicelist_get_selected_id(GtkIconView *iconview);

gchar *servicelist_service_short_name(gchar *name)
{
	gulong length;
	gchar buffer[256*4];
	gchar *shortname;
	
	if (name == NULL) return NULL;
	
	length = g_utf8_strlen((const gchar *)name, 256*4);
	if (length > 16) {
		memset(buffer, 0, sizeof(buffer));
		g_utf8_strncpy((gchar *)&buffer, (const gchar *)name, 16);
		shortname = g_strconcat((const gchar *)&buffer, "...", NULL);
	} else {
		shortname = g_strdup(name);
	}
	
	return shortname;
}

gboolean servicelist_preselect_service_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
	guint32 id;
	guchar type;
	
	gtk_tree_model_get(model, iter, COL_ID, &id, COL_TYPE, &type, -1);
	
	if ((id == current_service) && (type == current_service_type)) {
		gtk_icon_view_select_path(GTK_ICON_VIEW(servicelist.icon_view), path);
		gtk_icon_view_scroll_to_path(GTK_ICON_VIEW(servicelist.icon_view), path, FALSE, 0.0, 0.0);
		return TRUE;
	}
	
	return FALSE;
}

gboolean servicelist_count_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, guint32 *indexcount)
{
	guchar type;
	
	gtk_tree_model_get(model, iter, COL_TYPE, &type, -1);
	
	if (type != SERVICE_NETSHARE) *indexcount += 1;
	
	return FALSE;
}

gboolean servicelist_save_index_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, struct _sindex *index)
{
	guchar type;
	guint32 id;
	
	gtk_tree_model_get(model, iter, COL_ID, &id, COL_TYPE, &type, -1);
	
	if (type != SERVICE_NETSHARE) {
		index->vec[index->ptr] = id;
		index->ptr += 1;
	}
	
	return FALSE;
}

gboolean servicelist_close_window(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	struct _sindex index;
	
	index.count = 0;
	gtk_tree_model_foreach(GTK_TREE_MODEL(servicelist.store), (GtkTreeModelForeachFunc)servicelist_count_foreach, (gpointer)&index.count);
	
	index.vec = (guint32 *)g_malloc0(1*sizeof(guint32));
	
	if (index.count > 0) {
		index.vec = (guint32 *)g_realloc(index.vec, (index.count+1)*sizeof(guint32));
		index.ptr = 0;
		gtk_tree_model_foreach(GTK_TREE_MODEL(servicelist.store), (GtkTreeModelForeachFunc)servicelist_save_index_foreach, (gpointer)&index);
	}
	index.vec[index.count] = 0;
	index.count += 1;
	
	service_write_index(index.vec, index.count);
	
	g_free(index.vec);
	
	if (window) {
		gtk_widget_destroy(window);
		window = NULL;
	}
	
	return FALSE;
}

gboolean servicelist_remove_service_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, service *svc)
{
	guint id;
	
	gtk_tree_model_get(model, iter, COL_ID, &id, -1);
	
	if (id == svc->id) {
		gtk_list_store_remove(GTK_LIST_STORE(model), iter);
		
		return TRUE;
	} 
	
	return FALSE;
}

gboolean servicelist_remove_service_cb(service *svc)
{
	gtk_tree_model_foreach(GTK_TREE_MODEL(servicelist.store), (GtkTreeModelForeachFunc)servicelist_remove_service_foreach, (gpointer)svc);
	
	return TRUE;
}

gboolean servicelist_update_service_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, service *svc)
{
	gchar *name;
	guint id;
	GdkPixbuf *image;
	GError *error;
	gchar strid[15];
	gchar *imgfn, *tmpfn;
	const gchar *homedir;
	gchar *shortname;
	
	gtk_tree_model_get(model, iter, COL_ID, &id, -1);
	
	if (id == svc->id) {
		switch (svc->type) {
			case SERVICE_TV_RADIO: 
				imgfn = g_strdup(RESOURCE_SVCLIST_TVRADIO);
				break;
			case SERVICE_DATA: 
				imgfn = g_strdup(RESOURCE_SVCLIST_INTERNET);
				break;
			case SERVICE_TRANSPONDER: 
				imgfn = g_strdup(RESOURCE_SVCLIST_TRANSPONDER);
				break;
			case SERVICE_RADIO: 
				imgfn = g_strdup(RESOURCE_SVCLIST_RADIO);
				break;
			default:
				imgfn = g_strdup(RESOURCE_SVCLIST_TRANSPONDER);
				break;
		}
		
		if ((svc->picture != NULL) && (strlen(svc->picture) > 0)) {
			if ((homedir = g_getenv("HOME")) == NULL) homedir = g_get_home_dir();
			memset(strid, 0, sizeof(strid));
			g_sprintf(strid, "%u.png", svc->id);
			tmpfn = g_build_filename(homedir, strid, NULL);
			if (g_file_test(tmpfn, G_FILE_TEST_EXISTS)) imgfn = g_strdup(tmpfn);
			g_free(tmpfn);
		}
		
		error = NULL;
		image = gdk_pixbuf_new_from_file(imgfn, &error);
		g_free(imgfn);
		if (error != NULL) {
			g_warning(error->message);
			g_error_free(error);
		}
		
		shortname = servicelist_service_short_name(svc->name);
		name = g_markup_printf_escaped("<b>%s</b>\n<small>%i %s %s %i</small>", shortname, svc->freq, mainwindow_string_definition(DEFINITION_POLARISATION, svc->pol), mainwindow_string_definition(DEFINITION_FEC, svc->fec), svc->sr);
		gtk_list_store_set(GTK_LIST_STORE(model), iter, COL_TEXT, name, COL_PIXBUF, image, COL_TYPE, svc->type, -1);
		
		g_object_unref(image);
		g_free(name);
		g_free(shortname);
		
		return TRUE;
	} 
	
	return FALSE;
}

gboolean servicelist_update_service_cb(service *svc)
{
	gtk_tree_model_foreach(GTK_TREE_MODEL(servicelist.store), (GtkTreeModelForeachFunc)servicelist_update_service_foreach, (gpointer)svc);
	
	return TRUE;
}

gboolean servicelist_add_service_cb(service *svc)
{
	GtkTreeIter iter;
	gchar *name;
	GdkPixbuf *image;
	GError *error;
	gchar strid[15];
	gchar *imgfn, *tmpfn;
	const gchar *homedir;
	gchar *shortname;
	frontend_t frontend;
		
	switch (svc->type) {
		case SERVICE_TV_RADIO: 
			imgfn = g_strdup(RESOURCE_SVCLIST_TVRADIO);
			break;
		case SERVICE_DATA: 
			imgfn = g_strdup(RESOURCE_SVCLIST_INTERNET);
			break;
		case SERVICE_TRANSPONDER: 
			imgfn = g_strdup(RESOURCE_SVCLIST_TRANSPONDER);
			break;
		case SERVICE_RADIO: 
			imgfn = g_strdup(RESOURCE_SVCLIST_RADIO);
			break;
		default:
			imgfn = g_strdup(RESOURCE_SVCLIST_TRANSPONDER);
			break;
	}
	
	if ((svc->picture != NULL) && (strlen(svc->picture) > 0)) {
		if ((homedir = g_getenv("HOME")) == NULL) homedir = g_get_home_dir();
		memset(strid, 0, sizeof(strid));
		g_sprintf(strid, "%u.png", svc->id);
		tmpfn = g_build_filename(homedir, strid, NULL);
		if (g_file_test(tmpfn, G_FILE_TEST_EXISTS)) imgfn = g_strdup(tmpfn);
		g_free(tmpfn);
	}
	
	error = NULL;
	image = gdk_pixbuf_new_from_file(imgfn, &error);
	g_free(imgfn);
	if (error != NULL) {
		g_warning(error->message);
		g_error_free(error);
	}
	
	frontend = mainwindow_get_frontend();
	
	shortname = servicelist_service_short_name(svc->name);	
	
	if (dvb_calculate_values(frontend, svc, TRUE) == DVBSTATUS_VALUES_ALL_GOOD) {
		name = g_markup_printf_escaped("<b>%s</b>\n<small>%i %s %s %i</small>", shortname, svc->freq, mainwindow_string_definition(DEFINITION_POLARISATION, svc->pol), mainwindow_string_definition(DEFINITION_FEC, svc->fec), svc->sr);
	} else {
		name = g_markup_printf_escaped("<span foreground=\"grey\"><b>%s</b>\n<small>%i %s %s %i</small></span>", shortname, svc->freq, mainwindow_string_definition(DEFINITION_POLARISATION, svc->pol), mainwindow_string_definition(DEFINITION_FEC, svc->fec), svc->sr);
	}
		
	gtk_list_store_append(servicelist.store, &iter);
	gtk_list_store_set(servicelist.store, &iter, COL_TEXT, name, COL_PIXBUF, image, COL_ID, svc->id, COL_TYPE, svc->type, -1);
	
	g_object_unref(image);
	g_free(name);
	g_free(shortname);
	
	return TRUE;
}

gboolean servicelist_add_netservice_cb(netservice *nsvc)
{
	GtkTreeIter iter;
	gchar *name;
	GdkPixbuf *image;
	GError *error;
	gchar *imgfn;
	gchar *shortname;
	gchar *address;
	
	imgfn = g_strdup(RESOURCE_SVCLIST_NETSHARE);
	
	error = NULL;
	image = gdk_pixbuf_new_from_file(imgfn, &error);
	g_free(imgfn);
	if (error != NULL) {
		g_warning(error->message);
		g_error_free(error);
	}
	
	shortname = servicelist_service_short_name(nsvc->name);	
	address = mainwindow_get_client_net_address();
	name = g_markup_printf_escaped("<b>%s</b>\n<small>%s</small>", shortname, address); 
	gtk_list_store_append(servicelist.store, &iter);
	gtk_list_store_set(servicelist.store, &iter, COL_TEXT, name, COL_PIXBUF, image, COL_ID, nsvc->id, COL_TYPE, SERVICE_NETSHARE, -1);
	
	g_object_unref(image);
	g_free(name);
	g_free(address);
	g_free(shortname);
	
	return TRUE;
}


static void servicelist_item_activated(GtkIconView *icon_view, GtkTreePath *tree_path, gpointer data)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	guint32 id;
	guchar type;
	
	model = gtk_icon_view_get_model(icon_view);
	
	gtk_tree_model_get_iter(model, &iter, tree_path);
	gtk_tree_model_get(model, &iter, COL_ID, &id, COL_TYPE, &type, -1);
	
	current_service = id;
	current_service_type = type;
	
	mainwindow_tune_start(id, type);
	
	servicelist_close_window(NULL, NULL, NULL);
}

void servicelist_new_service(GtkMenuItem *menuitem, gpointer data)
{
	service_assistant_new(NULL, GTK_WINDOW(window));
}

void servicelist_edit_service(GtkMenuItem *menuitem, gpointer data)
{
	guint id = -1;
	service *svc;
	
	if ((id = servicelist_get_selected_id(GTK_ICON_VIEW(servicelist.icon_view))) != -1) {
		if ((svc = service_read(id)) != NULL) {
			service_assistant_new(svc, GTK_WINDOW(window));
		}
	}
	
}

void servicelist_remove_service(GtkMenuItem *menuitem, gpointer data)
{
	GtkWidget *dialog;
	guint id = -1;
	gint qres;
	service *svc;
	
	if ((id = servicelist_get_selected_id(GTK_ICON_VIEW(servicelist.icon_view))) != -1) {
		if ((svc = service_read(id)) != NULL) {
			dialog = gtk_message_dialog_new(GTK_WINDOW(window),
											GTK_DIALOG_DESTROY_WITH_PARENT,
											GTK_MESSAGE_WARNING,
											GTK_BUTTONS_YES_NO,
											_("Removing service \"%s\". Are you sure?"),
											svc->name, 
											NULL);
			qres = gtk_dialog_run(GTK_DIALOG(dialog));
			gtk_widget_destroy(dialog);
			
			if (qres == GTK_RESPONSE_YES) {
				service_remove(svc, servicelist_remove_service_cb);		
			}
		}
	}
	
}

static void servicelist_reload_net_services(GtkMenuItem *menuitem, gpointer data)
{
	struct _netgate *netgate;
			
	netgate = mainwindow_get_netgate();
	
	if (netgate->clientconnected) {
		mainwindow_network_servicelist_refresh();
		gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_netreload), FALSE);
	}
}

static void servicelist_clear_net_services(GtkMenuItem *menuitem, gpointer data)
{
	struct _netgate *netgate;
			
	netgate = mainwindow_get_netgate();
	
	if (netgate->clientconnected) {
		servicelist_remove_net_services();
		service_net_clear();
		if (service_net_get_number() == 0)
			gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_netclear), FALSE);
	}
}

gboolean servicelist_remove_net_services_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, GList **rowref_list)
{
	guchar type;
	GtkTreeRowReference *rowref;
	
	gtk_tree_model_get(model, iter, COL_TYPE, &type, -1);
	
	if (type == SERVICE_NETSHARE) {
		rowref = gtk_tree_row_reference_new(model, path);
		*rowref_list = g_list_append(*rowref_list, rowref);
	}
	
	return FALSE;
}

gboolean servicelist_remove_net_services(void)
{
	GtkTreeModel *model;
	GList *rr_list = NULL;
	GList *node;
	GtkTreePath *path;
	GtkTreeIter iter;
	
	model = gtk_icon_view_get_model(GTK_ICON_VIEW(servicelist.icon_view));
	gtk_tree_model_foreach(GTK_TREE_MODEL(model), (GtkTreeModelForeachFunc)servicelist_remove_net_services_foreach, &rr_list);
	
	for (node = rr_list;  node != NULL;  node = node->next) {
		if ((path = gtk_tree_row_reference_get_path((GtkTreeRowReference*)node->data)) != NULL) {
			if (gtk_tree_model_get_iter(GTK_TREE_MODEL(model), &iter, path)) {
				gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
			}
		}
	}
	
	g_list_foreach(rr_list, (GFunc)gtk_tree_row_reference_free, NULL);
	g_list_free(rr_list);
	
	return TRUE;
}

gboolean servicelist_reload_net_services_cb(gpointer data)
{
	if (window) {
		servicelist_remove_net_services();		
		service_net_read_foreach(servicelist_add_netservice_cb);
		gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_netreload), TRUE);
		if (service_net_get_number() > 0)
			gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_netclear), TRUE);
	}
	
	return FALSE;
}

static guint servicelist_get_selected_id(GtkIconView *iconview)
{
	GList *selitems;
	GtkTreePath *path;
	GtkTreeModel *model;
	GtkTreeIter iter;
	guint id = -1;
	guchar type = -1;
	
	if ((selitems = gtk_icon_view_get_selected_items(iconview)) != NULL) {
		path = (GtkTreePath *)g_list_nth_data(selitems, 0);
		model = gtk_icon_view_get_model(iconview);
		gtk_tree_model_get_iter(model, &iter, path);
		gtk_tree_model_get(model, &iter, COL_ID, &id, COL_TYPE, &type, -1);
		g_list_foreach(selitems, (GFunc)gtk_tree_path_free, NULL);
		g_list_free(selitems);
	}
	
	//Get only local services. If we have netshare, return -1
	if ((type == -1) || (type == SERVICE_NETSHARE)) id = -1;
	
	return id;
}

static void servicelist_popup_menu_show(GtkWidget *widget, GdkEventButton *event)
{
	gint button, event_time;
	
	if (servicelist_get_selected_id(GTK_ICON_VIEW(widget)) == -1) {
		gtk_widget_set_sensitive(servicelist.pm_edit, FALSE);
		gtk_widget_set_sensitive(servicelist.pm_delete, FALSE);
	} else {
		gtk_widget_set_sensitive(servicelist.pm_edit, TRUE);
		gtk_widget_set_sensitive(servicelist.pm_delete, TRUE);
	}
	
	if (event) {
		button = event->button;
		event_time = event->time;
	} else {
		button = 0;
		event_time = gtk_get_current_event_time();
	}
	
	
	gtk_menu_popup(GTK_MENU(servicelist.pm_menu), NULL, NULL, NULL, NULL, button, event_time);
}

static gboolean servicelist_button_press(GtkWidget *widget, GdkEventButton *event)
{
	if ((event->button == 3) && (event->type == GDK_BUTTON_PRESS)) {
		servicelist_popup_menu_show(widget, event);
		return TRUE;
    }
	
	return FALSE;
}

static gboolean servicelist_popup_menu(GtkWidget *widget)
{
	servicelist_popup_menu_show(widget, NULL);
	return TRUE;
}

void servicelist_selection_changed(GtkIconView *iconview, gpointer data)
{
	if (servicelist_get_selected_id(iconview) == -1) {
		gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_edit), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_remove), FALSE);
	} else {
		gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_edit), TRUE);
		gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_remove), TRUE);
	}
	
}

gboolean servicelist_new(GtkWindow *rootwindow)
{
	GtkWidget *listscroll;
    GtkCellRenderer *renderer;
	GtkWidget *vbox;
	struct _netgate *netgate;
		
	netgate = mainwindow_get_netgate();
		
	if (!window) {
		window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_set_transient_for(GTK_WINDOW(window), GTK_WINDOW(rootwindow));
		gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER_ON_PARENT);
		gtk_window_set_modal(GTK_WINDOW(window), TRUE);
		gtk_window_set_title(GTK_WINDOW (window), _("Service browser"));
		gtk_window_set_default_size(GTK_WINDOW(window), 500, 300);
		
		g_signal_connect(window, "destroy", G_CALLBACK(servicelist_close_window), NULL);
		
		vbox = gtk_vbox_new(FALSE, 0);
		
		servicelist.toolbar = gtk_toolbar_new();
		
		servicelist.tb_new = gtk_tool_button_new_from_stock(GTK_STOCK_NEW);
		gtk_widget_set_tooltip_text(GTK_WIDGET(servicelist.tb_new), _("Add new service"));
		g_object_set(servicelist.tb_new, "is-important", TRUE, NULL);
		servicelist.tb_edit = gtk_tool_button_new_from_stock(GTK_STOCK_EDIT);
		gtk_widget_set_tooltip_text(GTK_WIDGET(servicelist.tb_edit), _("Edit service parameters"));
		g_object_set(servicelist.tb_edit, "is-important", TRUE, NULL);
		servicelist.tb_remove = gtk_tool_button_new_from_stock(GTK_STOCK_REMOVE);
		gtk_widget_set_tooltip_text(GTK_WIDGET(servicelist.tb_remove), _("Remove service"));
		g_object_set(servicelist.tb_remove, "is-important", TRUE, NULL);
		servicelist.tb_separator = gtk_separator_tool_item_new();
		servicelist.tb_netreload = gtk_tool_button_new_from_stock(GTK_STOCK_REFRESH);
		gtk_widget_set_tooltip_text(GTK_WIDGET(servicelist.tb_netreload), _("Reload shared services list"));
		//g_object_set(servicelist.tb_netreload, "is-important", TRUE, NULL);
		servicelist.tb_netclear = gtk_tool_button_new_from_stock(GTK_STOCK_CLEAR);
		gtk_widget_set_tooltip_text(GTK_WIDGET(servicelist.tb_netclear), _("Clear shared services"));
		//g_object_set(servicelist.tb_netclear, "is-important", TRUE, NULL);
		
		//servicelist.tb_connect = gtk_toggle_tool_button_new_from_stock(GTK_STOCK_CONNECT);
						
		gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_edit), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_remove), FALSE);
		
		if (!netgate->clientconnected)
			gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_netreload), FALSE);
		
		if ((!netgate->clientconnected) || (service_net_get_number() == 0))
			gtk_widget_set_sensitive(GTK_WIDGET(servicelist.tb_netclear), FALSE);
		
		gtk_toolbar_set_show_arrow(GTK_TOOLBAR(servicelist.toolbar), TRUE);
		gtk_toolbar_set_style(GTK_TOOLBAR(servicelist.toolbar), GTK_TOOLBAR_BOTH_HORIZ);
		
		gtk_toolbar_insert(GTK_TOOLBAR(servicelist.toolbar), servicelist.tb_new, 0);
		gtk_toolbar_insert(GTK_TOOLBAR(servicelist.toolbar), servicelist.tb_edit, 1);
		gtk_toolbar_insert(GTK_TOOLBAR(servicelist.toolbar), servicelist.tb_remove, 2);
		gtk_toolbar_insert(GTK_TOOLBAR(servicelist.toolbar), servicelist.tb_separator, 3);
		gtk_toolbar_insert(GTK_TOOLBAR(servicelist.toolbar), servicelist.tb_netreload, 4);
		gtk_toolbar_insert(GTK_TOOLBAR(servicelist.toolbar), servicelist.tb_netclear, 5);
		
		gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(servicelist.toolbar), FALSE, TRUE, 0);
		
		servicelist.store = gtk_list_store_new(NUM_COLS, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_UINT, G_TYPE_UCHAR);
		gtk_list_store_clear(servicelist.store);
		//TODO: add callback
		service_read_foreach(servicelist_add_service_cb, servicelist_add_netservice_cb, NULL);
		
		servicelist.icon_view = gtk_icon_view_new_with_model(GTK_TREE_MODEL(servicelist.store));
		g_signal_connect(servicelist.icon_view, "item-activated", G_CALLBACK(servicelist_item_activated), NULL);
		g_signal_connect(servicelist.icon_view, "button-press-event", G_CALLBACK(servicelist_button_press), NULL);
		g_signal_connect(servicelist.icon_view, "popup-menu", G_CALLBACK(servicelist_popup_menu), NULL);
		g_signal_connect(servicelist.icon_view, "selection-changed", G_CALLBACK(servicelist_selection_changed), NULL);
		g_object_unref(servicelist.store);
		
		//if (current_service != 0) gtk_tree_model_foreach(GTK_TREE_MODEL(servicelist.store), servicelist_preselect_service_foreach, NULL);
		
		g_signal_connect_swapped(G_OBJECT(servicelist.tb_new), "clicked", G_CALLBACK(servicelist_new_service), NULL);
		g_signal_connect_swapped(G_OBJECT(servicelist.tb_edit), "clicked", G_CALLBACK(servicelist_edit_service), NULL);
		g_signal_connect_swapped(G_OBJECT(servicelist.tb_remove), "clicked", G_CALLBACK(servicelist_remove_service), NULL);
		g_signal_connect_swapped(G_OBJECT(servicelist.tb_netreload), "clicked", G_CALLBACK(servicelist_reload_net_services), NULL);
		g_signal_connect_swapped(G_OBJECT(servicelist.tb_netclear), "clicked", G_CALLBACK(servicelist_clear_net_services), NULL);
				
		//g_signal_connect_swapped(G_OBJECT(selectall), "clicked", G_CALLBACK(select_all), entry);
		
		servicelist.pm_menu = gtk_menu_new();
		servicelist.pm_new = gtk_image_menu_item_new_from_stock(GTK_STOCK_NEW, NULL);
		servicelist.pm_edit = gtk_image_menu_item_new_from_stock(GTK_STOCK_EDIT, NULL);
		servicelist.pm_delete = gtk_image_menu_item_new_from_stock(GTK_STOCK_DELETE, NULL);
		
		servicelist.pm_group = gtk_accel_group_new();
		gtk_window_add_accel_group(GTK_WINDOW(window), servicelist.pm_group);
		gtk_menu_set_accel_group(GTK_MENU(servicelist.pm_menu), servicelist.pm_group);
		
		#if GTK_CHECK_VERSION(2, 22, 0)
			gtk_widget_add_accelerator(servicelist.pm_new, "activate", servicelist.pm_group, GDK_KEY_N, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
			gtk_widget_add_accelerator(servicelist.pm_edit, "activate", servicelist.pm_group, GDK_KEY_E, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
			gtk_widget_add_accelerator(servicelist.pm_delete, "activate", servicelist.pm_group, GDK_KEY_D, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
		#else
			gtk_widget_add_accelerator(servicelist.pm_new, "activate", servicelist.pm_group, GDK_N, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
			gtk_widget_add_accelerator(servicelist.pm_edit, "activate", servicelist.pm_group, GDK_E, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
			gtk_widget_add_accelerator(servicelist.pm_delete, "activate", servicelist.pm_group, GDK_D, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
		#endif
		
		g_signal_connect(G_OBJECT(servicelist.pm_new), "activate", G_CALLBACK(servicelist_new_service), NULL);
		g_signal_connect(G_OBJECT(servicelist.pm_edit), "activate", G_CALLBACK(servicelist_edit_service), NULL);
		g_signal_connect(G_OBJECT(servicelist.pm_delete), "activate", G_CALLBACK(servicelist_remove_service), NULL);
		
		gtk_menu_shell_append(GTK_MENU_SHELL(servicelist.pm_menu), servicelist.pm_new);
		gtk_menu_shell_append(GTK_MENU_SHELL(servicelist.pm_menu), servicelist.pm_edit);
		gtk_menu_shell_append(GTK_MENU_SHELL(servicelist.pm_menu), servicelist.pm_delete);
		
		gtk_widget_show_all(servicelist.pm_menu);
		gtk_menu_attach_to_widget(GTK_MENU(servicelist.pm_menu), servicelist.icon_view, NULL);
		
		gtk_icon_view_set_selection_mode(GTK_ICON_VIEW (servicelist.icon_view), GTK_SELECTION_SINGLE);
		
		#if GTK_CHECK_VERSION(2, 22, 0)
			gtk_icon_view_set_item_orientation(GTK_ICON_VIEW(servicelist.icon_view), GTK_ORIENTATION_HORIZONTAL);
		#else
			gtk_icon_view_set_orientation(GTK_ICON_VIEW(servicelist.icon_view), GTK_ORIENTATION_HORIZONTAL);
		#endif
		//gtk_icon_view_set_columns(GTK_ICON_VIEW(servicelist.icon_view), 2);
		gtk_icon_view_set_reorderable(GTK_ICON_VIEW (servicelist.icon_view), TRUE);
		
		renderer = gtk_cell_renderer_pixbuf_new();
		gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(servicelist.icon_view), renderer, TRUE);
		gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(servicelist.icon_view), renderer, "pixbuf", COL_PIXBUF, NULL);
		
		renderer = gtk_cell_renderer_text_new();
		gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(servicelist.icon_view), renderer, TRUE);
		gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(servicelist.icon_view), renderer, "markup", COL_TEXT, NULL);
		
		listscroll = gtk_scrolled_window_new(NULL, NULL);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(listscroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(listscroll), GTK_SHADOW_IN);
		gtk_container_set_border_width(GTK_CONTAINER(listscroll), 5);
		gtk_container_add(GTK_CONTAINER(listscroll), servicelist.icon_view);
		
		gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(listscroll), TRUE, TRUE, 0);
		
		gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(vbox));
		
		if (current_service != 0) 
			gtk_tree_model_foreach(GTK_TREE_MODEL(servicelist.store), servicelist_preselect_service_foreach, NULL);
		
		gtk_widget_grab_focus(servicelist.icon_view);
	}
	
	#if GTK_CHECK_VERSION(2, 18, 0)
		if (!gtk_widget_get_visible(window)) {
	#else
		if (!GTK_WIDGET_VISIBLE(window)) {
	#endif
		gtk_widget_show_all(window);
	} else {
		gtk_widget_destroy(window);
		window = NULL;
    }
	
	return TRUE;
}
