/*
 *      dvbscan.c
 *      
 *      Copyright 2009 - 2010 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib.h>
#include <stdio.h>
#include <stdint.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/ioctl.h>
#include <time.h>
#include <linux/dvb/dmx.h>
#include "dvbscan.h"
#include "common/iso3166.h"

static GHashTable *scansvchash = NULL;
static GSList *netsearch = NULL;
static gchar *networkname = NULL;
static struct tm *timestamp = NULL;


GHashTable *dvbscan_get_services(void)
{
	return scansvchash;
}

GSList *dvbscan_get_netsearch(void)
{
	return netsearch;
}

gchar *dvbscan_get_network_name(void)
{
	return networkname;
}

struct tm *dvbscan_get_tptime(void)
{
	return timestamp;
}

static int start_filter(guint16 pid, guint8 tid, guint adapter, guint demux)
{
	struct dmx_sct_filter_params f;
	gint fd;
	gchar *demuxfn;
	
	demuxfn = g_strdup_printf("/dev/dvb/adapter%u/demux%u", adapter, demux, NULL);
	
	if ((fd = open (demuxfn, O_RDWR | O_NONBLOCK)) < 0) {
		g_free(demuxfn);
		goto err0;
	}
	
	g_free(demuxfn);
	
	memset(&f, 0, sizeof(f));

	f.pid = pid;
	f.filter.filter[0] = tid;
	f.filter.mask[0]   = 0xff;
	

	f.timeout = 0;
	f.flags = DMX_IMMEDIATE_START | DMX_CHECK_CRC;

	if (ioctl(fd, DMX_SET_FILTER, &f) == -1) {
		printf("ioctl DMX_SET_FILTER failed\n");
		goto err1;
	}

	return fd;

err1:
	ioctl (fd, DMX_STOP);
	close (fd);
err0:
	return -1;
}

static gint stop_filter(gint fd)
{
	ioctl (fd, DMX_STOP);
	close (fd);
}

static gint read_filter(gint filterfd, gchar *buffer, gint bufsize, guchar timeout)
{
	fd_set readset;
	struct timeval tv;
	gint readb = 0;
	
	FD_ZERO(&readset);
	FD_SET(filterfd, &readset);
	
	tv.tv_sec = timeout;
	tv.tv_usec = 0;
	select(filterfd+1, &readset, NULL, NULL, timeout == 0 ? NULL : &tv);
	
	if (FD_ISSET(filterfd, &readset)) {
		readb = read(filterfd, buffer, bufsize);
	}
		
	return readb;
}

gboolean dvbscan_free_list(void)
{
	if (scansvchash != NULL) {
		g_hash_table_destroy(scansvchash);
		scansvchash = NULL;
	}
		
	if (netsearch != NULL) {
		g_slist_free(netsearch);
		netsearch = NULL;
	}
	
	if (networkname != NULL) {
		g_free(networkname);
		networkname = NULL;
	}
	
	if (timestamp != NULL) {
		g_free(timestamp);
		timestamp = NULL;
	}
	
	return TRUE;
}

static gchar *dvbscan_convert_encoding(gchar *string)
{
	gchar *encoding;
	const gchar *utf = "UTF-8";
	gsize bread, bwritten;
	gboolean needoffset = TRUE;
	
	switch (string[0]) {
		case 0x01: encoding = "ISO8859-5"; break;
		case 0x02: encoding = "ISO8859-6"; break;
		case 0x03: encoding = "ISO8859-7"; break;
		case 0x04: encoding = "ISO8859-8"; break;
		case 0x05: encoding = "ISO8859-9"; break;
		case 0x06: encoding = "ISO8859-10"; break;
		case 0x07: encoding = "ISO8859-11"; break;
		case 0x08: encoding = "ISO8859-12"; break;
		case 0x09: encoding = "ISO8859-13"; break;
		case 0x0A: encoding = "ISO8859-14"; break;
		case 0x0B: encoding = "ISO8859-15"; break;
		default:   encoding = "ISO_6937";  needoffset = FALSE; break;
	}
	
	return g_convert(needoffset ? string+1 : string, -1, utf, encoding, &bread, &bwritten, NULL);
}

gchar *dvbscan_get_cas_name(guint caid) 
{
	gint i = 0;
	gboolean found = FALSE;
	gchar *res;
	
	typedef struct {
		guint caid;
		const gchar *str;
	} enc_table;
	
	enc_table enctable[] = {
		{0x0000, "Reserved"},
		{0x0100, "Seca/MediaGuard"},
		{0x0500, "Viaccess"},
		{0x0600, "Irdeto"},
		{0x0634, "Irdeto 2M"},
		{0x0652, "Irdeto 2"},
		{0x0900, "Videoguard"},
		{0x090F, "Videoguard"},
		{0x0B00, "Conax"},
		{0x0C00, "NTL"},
		{0x0D00, "Cryptoworks"},
		{0x0E00, "Power VU"},
		{0x1700, "Betacrypt"},
		{0x1800, "Nagravision"},
		{0x22e3, "Codicrypt"},
		{0x2600, "BISS"},
		{0x4A20, "AlphaCrypt"},
		{0x4A70, "DreamCrypt"},
		{0x4A80, "THALESCrypt"},
		{0xA101, "Rosscrypt"},
		{0,      NULL}
	};
	
	while (enctable[i].str != NULL) {
		if (caid == enctable[i].caid) {
			return g_strdup(enctable[i].str);
		}		
		i++;
	}
	
	return g_strdup_printf("Unknown 0x%x", caid); 
}


void dvbscan_free_hash(gpointer data)
{
	struct _srcservice *scansvc;
	
	scansvc = (struct _srcservice *)data;
	if (scansvc->name != NULL) g_free(scansvc->name);
	if (scansvc->provider != NULL) g_free(scansvc->provider);
	g_free(scansvc);
}

void dvbscan_add_caid(struct _srcservice *scansvc, gushort caid)
{
	gint i, freeindex;
	gboolean idexists;
	
	freeindex = -1;
	idexists = FALSE;
	
	if ((caid == 0) || (scansvc == NULL)) return;
	
	for (i=0; i<16; i++) {
		if ((freeindex == -1) && (scansvc->caid[i] == 0)) freeindex = i;
		if (caid == scansvc->caid[i]) {
			idexists = TRUE;
			break;
		}
	}
	
	if ((!idexists) && (freeindex != -1)) {
		scansvc->caid[freeindex] = caid;
		printf(" CAID: 0x%x\n", caid);
	}
}

gchar dvbscan_add_audio_pid(struct _srcservice *scansvc, gushort audiopid)
{
	gint i, freeindex;
	gboolean idexists;
	
	freeindex = -1;
	idexists = FALSE;
	
	if ((audiopid == 0) || (scansvc == NULL)) return -1;
	
	for (i=0; i<16; i++) {
		if ((freeindex == -1) && (scansvc->apid[i] == 0)) freeindex = i;
		if (audiopid == scansvc->apid[i]) {
			idexists = TRUE;
			break;
		}
	}
	
	if ((!idexists) && (freeindex != -1)) {
		scansvc->apid[freeindex] = audiopid;
		printf(" Audio PID: 0x%x\n", audiopid);
	}
	
	return (guchar)freeindex;
}

gchar dvbscan_add_data_pid(struct _srcservice *scansvc, gushort datapid)
{
	gint i, freeindex;
	gboolean idexists;
	
	freeindex = -1;
	idexists = FALSE;
	
	if ((datapid == 0) || (scansvc == NULL)) return -1;
	
	for (i=0; i<16; i++) {
		if ((freeindex == -1) && (scansvc->dpid[i] == 0)) freeindex = i;
		if (datapid == scansvc->dpid[i]) {
			idexists = TRUE;
			break;
		}
	}
	
	if ((!idexists) && (freeindex != -1)) {
		scansvc->dpid[freeindex] = datapid;
		printf(" Data PID: 0x%x\n", datapid);
	}
	
	return (guchar)freeindex;
}

void dvbscan_add_audio_language(struct _srcservice *scansvc, gchar *buffer, gchar langid)
{
	gchar langbuf[4];
	gushort langcode;
	
	if ((langid < 0) || (langid > DVBSCAN_SERVICE_LANG_ID) || (scansvc == NULL) || (buffer == NULL)) return;
	
	memset(langbuf, 0, sizeof(langbuf));
	memcpy(&langbuf, buffer, 3);
	
	if (langid == DVBSCAN_SERVICE_LANG_ID) {
		if ((langcode = iso3166_get_code_from_a3((const char *)buffer)) != 0) {
			scansvc->svclang = langcode;
			printf(" Service language: %u (%s)\n", langcode, langbuf);
		}
	} else {
		if ((langcode = iso3166_get_code_from_a3((const char *)buffer)) != 0) {
			scansvc->lang[langid] = langcode;
			printf(" Language: %u (%s)\n", langcode, langbuf);
		}
	}
}

gint dvbscan_start(guint adapter, guint demux, void (status_cb)(gint status))
{
	gint   patfilter, pmtfilter, sdtfilter, nitfilter, tdtfilter;
	guchar buffer[4096];
	guchar table_id;
	guint  table_length;
	guint  transport_stream_id;
	guchar version_number;
	guchar section_number;
	guchar last_section_number;
	guint  program_info_length;
	guchar stream_type;
	gushort  caid;
	guint  elementary_PID;
	guint  ES_info_length;
	guchar descriptor_tag;
	guchar descriptor_length;
	
	guint  num_services;	
	gint   i,i2;
	gint   ptr = 0;
	gint   descriptorptr = 0;
	gchar  audioid;
	
	guint  original_network_id;
	gint   service_id;
	guchar service_status;
	guchar free_CA_mode;
	guint  descriptors_loop_length;
	guchar service_provider_name_length;
	guchar service_name_length;
	gchar  strbuf[256];
	
	guint  network_id;
	guint  network_descriptors_length;
	guchar service_type;
	guint  transport_stream_loop_length;
	guint  transport_descriptors_length;
		
	guint  y, m, d, k;
	guint  MJD_time;
	guint  UTC_time;
		
	GHashTableIter iter;
	gpointer key, value;
	
	struct _srcservice *scansvc;
	struct _netentry *netentry;
	
	//PAT SECTION
	if (status_cb != NULL) status_cb(DVBSCAN_STATUS_PAT);
		
	patfilter = start_filter(0x0000, 0x00, adapter, demux);
	
	if (read_filter(patfilter, (char *)&buffer, sizeof(buffer), DVBSCAN_PAT_READ_TIMEOUT) <= 0) {
		printf("Failed to read PAT table\n");
		return 0;
	}
	
	stop_filter(patfilter);
	
	printf("PAT section\n");
	
	table_id = buffer[0];
	table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
	transport_stream_id = (((buffer[3] & 0xff) << 8) | buffer[4]);
	version_number = (buffer[5] & 0x3e);
	section_number = buffer[6];
	last_section_number = buffer[7];
	
	num_services = ((table_length + 3 - DVBSCAN_SECTION_HEADER_LENGTH - DVBSCAN_SECTION_CRC_LENGTH) / 4);	
		
	if ((scansvchash != NULL) || (netsearch != NULL)) {
		dvbscan_free_list();
	}
	
	scansvchash = g_hash_table_new_full(g_int_hash, g_int_equal, NULL, (GDestroyNotify)dvbscan_free_hash);
	
	for (i=0; i<num_services; i++) {
		scansvc = g_malloc0(sizeof(struct _srcservice));
		scansvc->sid = (((buffer[8+(4*i)] & 0xff) << 8) | buffer[9+(4*i)]);
		scansvc->pmt = (((buffer[10+(4*i)] & 0x1f) << 8) | buffer[11+(4*i)]);
		scansvc->name = g_strdup_printf("Service %u", scansvc->sid);
		scansvc->provider = g_strdup_printf("Provider %u", scansvc->sid);
		scansvc->sticky = TRUE;
		if (scansvc->sid == 0) {
			g_free(scansvc);
		} else {
			g_hash_table_insert(scansvchash, (gpointer)&(scansvc->sid), (gpointer)scansvc);
			printf(" %u. SID = %u, PID = 0x%x\n", i, scansvc->sid, scansvc->pmt);
		}
	}
	
	printf("Read %i services from PAT\n", num_services-1);
		
	
	//PMT SECTION
	if (status_cb != NULL) status_cb(DVBSCAN_STATUS_PMT);
		
	i = 0;
	g_hash_table_iter_init(&iter, scansvchash);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		scansvc = (struct _srcservice *)value;
		
		if ((scansvc == NULL) || (scansvc->sid == 0)) { i++; continue; }
		
		pmtfilter = start_filter(scansvc->pmt, 0x02, adapter, demux);
		
		if (read_filter(pmtfilter, (char *)&buffer, sizeof(buffer), DVBSCAN_PMT_READ_TIMEOUT) <= 0) {
			printf("Failed to read PMT table\n");
			i++;
			continue;
		}
		
		stop_filter(pmtfilter);
		
		printf(" %i. PMT section PID 0x%x\n", i, scansvc->pmt);
		
		table_id = buffer[0];
		table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
		transport_stream_id = (((buffer[3] & 0xff) << 8) | buffer[4]);
		version_number = (buffer[5] & 0x3e);
		section_number = buffer[6];
		last_section_number = buffer[7];
		scansvc->pcr = (((buffer[8] & 0x1f) << 8) | buffer[9]);
		program_info_length = (((buffer[10] & 0x0f) << 8) | buffer[11]);
		
		ptr = 12;
		descriptorptr = 0;
		
		while ((program_info_length > 0) && (descriptorptr < program_info_length-1)) {
			descriptor_tag = buffer[ptr];
			descriptor_length = buffer[ptr+1] + 2;
			if (descriptor_tag == 0x0a) {
				dvbscan_add_audio_language(scansvc, &buffer[ptr+2], DVBSCAN_SERVICE_LANG_ID);
			} else if (descriptor_tag == 0x09) {
				caid = (buffer[ptr+2] << 8) | buffer[ptr+3];
				dvbscan_add_caid(scansvc, caid);
			} /*else printf("1. Descriptorlength: %i, tag 0x%x, length %i\n", ES_info_length, descriptor_tag, descriptor_length);*/
			ptr += descriptor_length;
			descriptorptr += descriptor_length;
		}
		
		while (ptr < table_length-1) {
			stream_type = buffer[ptr];
			elementary_PID = (((buffer[ptr+1] & 0x1f) << 8) | buffer[ptr+2]);
			ES_info_length = (((buffer[ptr+3] & 0x0f) << 8) | buffer[ptr+4]);
			if ((stream_type == 0x01) || (stream_type == 0x02) || (stream_type == 0x1b)) {
				scansvc->vpid = elementary_PID;
				printf(" Video PID: 0x%x\n", elementary_PID);
			} else if ((stream_type == 0x03) || (stream_type == 0x04) || (stream_type == 0x81)) {
				audioid = dvbscan_add_audio_pid(scansvc, elementary_PID);
			} else if (stream_type == 0x06) {
				dvbscan_add_data_pid(scansvc, elementary_PID);
			} /*else printf("Other: 0x%x\n", stream_type);*/

			if (ES_info_length > 3) {
				descriptorptr = 0;
				while (descriptorptr < ES_info_length-1) {
					descriptor_tag = buffer[ptr+5];
					descriptor_length = buffer[ptr+6] + 2;
					if (descriptor_tag == 0x0a) {
						dvbscan_add_audio_language(scansvc, &buffer[ptr+7], audioid);
					} else if (descriptor_tag == 0x09) {
						caid = (buffer[ptr+7] << 8) | buffer[ptr+8];
						dvbscan_add_caid(scansvc, caid);
					} else if ((stream_type == 0x06) && (descriptor_tag == 0x56)) {
						scansvc->ttpid = elementary_PID;
						printf(" Teletext PID: 0x%x\n", elementary_PID);
					} else if ((stream_type == 0x06) && (descriptor_tag == 0x59)) {
						scansvc->stpid = elementary_PID;
						printf(" Subtitles PID: 0x%x\n", elementary_PID);
					} else if ((stream_type == 0x06) && (descriptor_tag == 0x6a)) {
						scansvc->ac3pid = elementary_PID;
						printf(" AC3 Audio PID: 0x%x\n", elementary_PID);
					} /*else printf("2. Descriptorlength: %i, tag 0x%x, length %i\n", ES_info_length, descriptor_tag, descriptor_length);*/
					descriptorptr += descriptor_length;
				}
			}
			
			ptr += ES_info_length + 5;
		}
		
		if (scansvc->caid[0] != 0) scansvc->sticky = FALSE;
		i++;
	}
			
	//SDT SECTION
	if (status_cb != NULL) status_cb(DVBSCAN_STATUS_SDT);
	
	sdtfilter = start_filter(0x0011, 0x42, adapter, demux);
	
	if (read_filter(sdtfilter, (char *)&buffer, sizeof(buffer), DVBSCAN_SDT_READ_TIMEOUT) <= 0) {
		printf("Failed to read SDT table\n");
		return 2;
	}
		
	stop_filter(sdtfilter);
		
	printf("SDT section\n");
	
	table_id = buffer[0];
	table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
	transport_stream_id = (((buffer[3] & 0xff) << 8) | buffer[4]);
	version_number = (buffer[5] & 0x3e);
	section_number = buffer[6];
	last_section_number = buffer[7];
	original_network_id = (((buffer[8] & 0xff) << 8) | buffer[9]);
	
	ptr = 11;
	i = 0;
	
	while (ptr < (table_length-1)) {
		service_id = (((buffer[ptr+0] & 0xff) << 8) | buffer[ptr+1]);
		service_status = ((buffer[ptr+2] & 0xe0) >> 5);
		free_CA_mode = ((buffer[ptr+2] & 0x10) >> 4);
		descriptors_loop_length = (((buffer[ptr+3] & 0x0f) << 8) | buffer[ptr+4]);
		descriptorptr = 0;
		if ((scansvc = g_hash_table_lookup(scansvchash, &service_id)) != NULL) {
			scansvc->status = service_status;
			while (descriptorptr < (descriptors_loop_length - 1)) {
				descriptor_tag = buffer[ptr+descriptorptr+5];
				descriptor_length = buffer[ptr+descriptorptr+6]+2;
				if (descriptor_tag == 0x48) {
					scansvc->type = buffer[ptr+descriptorptr+7];
					service_provider_name_length = buffer[ptr+descriptorptr+8];
					memset(strbuf, 0, sizeof(strbuf));
					memcpy(strbuf, &buffer[ptr+descriptorptr+9], service_provider_name_length);
					strbuf[service_provider_name_length] = '\0';
					if (scansvc->provider != NULL) g_free(scansvc->provider);
					scansvc->provider = dvbscan_convert_encoding((gchar *)&strbuf);
					service_name_length = buffer[service_provider_name_length+ptr+descriptorptr+9];
					memset(strbuf, 0, sizeof(strbuf));
					memcpy(strbuf, &buffer[service_provider_name_length+ptr+descriptorptr+10], service_name_length);
					strbuf[service_name_length] = '\0';
					if (scansvc->name != NULL) g_free(scansvc->name);
					scansvc->name = dvbscan_convert_encoding((gchar *)&strbuf);
				} else if (descriptor_tag == 0x53) {
					for (i=0; i<((descriptor_length-2)/2); i++) {
						caid = (((buffer[ptr+descriptorptr+7] & 0xff) << 8) | buffer[ptr+descriptorptr+8]);
						dvbscan_add_caid(scansvc, caid);
					}
				} /*else printf("3. descriptor_length = %i (0x%x) loop %i\n", descriptor_length, descriptor_tag, descriptors_loop_length);*/
				descriptorptr += descriptor_length;
			}
			printf(" %u. %s - %s\n", i+1, scansvc->provider, scansvc->name);
			i++;
		}
		ptr += 5 + descriptors_loop_length;
	}
		
	
	//NIT SECTION
	if (status_cb != NULL) status_cb(DVBSCAN_STATUS_NIT);
	
	nitfilter = start_filter(0x10, 0x40, adapter, demux);
		
	if (read_filter(nitfilter, (char *)&buffer, sizeof(buffer), DVBSCAN_NIT_READ_TIMEOUT) <= 0) {
		printf("Failed to read NIT table\n");
		return 3;
	}
				
	stop_filter(nitfilter);
	
	printf("NIT section\n");
	
	table_id = buffer[0];
	table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
	network_id = (((buffer[3] & 0xff) << 8) | buffer[4]);
	version_number = (buffer[5] & 0x3e);
	section_number = buffer[6];
	last_section_number = buffer[7];
	network_descriptors_length = (((buffer[8] & 0x0f) << 8) | buffer[9]);
	
	ptr = 10;
	descriptorptr = 0;
	
	while ((network_descriptors_length > 0) && (descriptorptr < network_descriptors_length-1)) {
		descriptor_tag = buffer[ptr];
		descriptor_length = buffer[ptr+1] + 2;
		if (descriptor_tag == 0x40) {
			memset(strbuf, 0, sizeof(strbuf));
			memcpy(strbuf, &buffer[ptr+2], (descriptor_length-2)>255 ? 255 : (descriptor_length-2));
			networkname = dvbscan_convert_encoding((gchar *)&strbuf);
			printf("Network name: %s\n", networkname);
		} /*else printf("4. descriptor_tag = 0x%x, descriptor_length = %i\n", descriptor_tag, descriptor_length);*/
		/*else if (descriptor_tag == 0x09) {
			caid = (buffer[ptr+2] << 8) | buffer[ptr+3];
			printf("CAID: 0x%x\n", caid);
		}*/
		ptr += descriptor_length;
		descriptorptr += descriptor_length;
	}
	
	transport_stream_loop_length = (((buffer[ptr] & 0x1f) << 8) | buffer[ptr+1]);
	ptr = ptr + 2;
	
	while (ptr < table_length-1) {
		transport_stream_id = (((buffer[ptr] & 0x1f) << 8) | buffer[ptr+1]);
		original_network_id = (((buffer[ptr+2] & 0x1f) << 8) | buffer[ptr+3]);
		transport_descriptors_length = (((buffer[ptr+4] & 0x0f) << 8) | buffer[ptr+5]);
		
		if (transport_descriptors_length > 0) {
			descriptorptr = 0;
			while (descriptorptr < transport_descriptors_length-1) {
				descriptor_tag = buffer[ptr+descriptorptr+6];
				descriptor_length = buffer[ptr+descriptorptr+7] + 2;
				if (descriptor_tag == 0x41) {
					service_id = (((buffer[ptr+descriptorptr+8] & 0x1f) << 8) | buffer[ptr+descriptorptr+9]);
					service_type = buffer[ptr+descriptorptr+10];
					if ((scansvc = g_hash_table_lookup(scansvchash, &service_id)) != NULL) {
						scansvc->type = service_type;
					}
					printf("SID: 0x%x, Service type: 0x%x\n", service_id, service_type);
				} else if (descriptor_tag == 0x43) {
					netentry = g_malloc0(sizeof(struct _netentry));
					//memset(netentry, 0, sizeof(snetentry));
					netentry->frequency = 10*(((buffer[ptr+descriptorptr+8]  >> 4) & 0x0f) * 10000000 + (buffer[ptr+descriptorptr+8] & 0x0f) * 1000000 +
					                          ((buffer[ptr+descriptorptr+9]  >> 4) & 0x0f) * 100000   + (buffer[ptr+descriptorptr+9] & 0x0f) * 10000 +
								              ((buffer[ptr+descriptorptr+10] >> 4) & 0x0f) * 1000     + (buffer[ptr+descriptorptr+10] & 0x0f) * 100 +
								              ((buffer[ptr+descriptorptr+11] >> 4) & 0x0f) * 10       + (buffer[ptr+descriptorptr+11] & 0x0f));
					netentry->position =     (((buffer[ptr+descriptorptr+12] >> 4) & 0x0f) * 1000     + (buffer[ptr+descriptorptr+12] & 0x0f) * 100 +
								              ((buffer[ptr+descriptorptr+13] >> 4) & 0x0f) * 10       + (buffer[ptr+descriptorptr+13] & 0x0f));
					if (!(buffer[ptr+descriptorptr+14] >> 7)) netentry->position = -(netentry->position);
					netentry->polarization =   (buffer[ptr+descriptorptr+14] >> 5) & 0x03;
					netentry->modulation =      buffer[ptr+descriptorptr+14] & 0x1f;
					netentry->sr =        10*(((buffer[ptr+descriptorptr+15] >> 4) & 0x0f) * 10000000 + (buffer[ptr+descriptorptr+15] & 0x0f) * 1000000 +
					                          ((buffer[ptr+descriptorptr+16] >> 4) & 0x0f) * 100000   + (buffer[ptr+descriptorptr+16] & 0x0f) * 10000 +
								              ((buffer[ptr+descriptorptr+17] >> 4) & 0x0f) * 1000     + (buffer[ptr+descriptorptr+17] & 0x0f) * 100 +
								             (((buffer[ptr+descriptorptr+18] & 0xf0) >> 4) & 0x0f) * 10 + ((buffer[ptr+descriptorptr+18] & 0xf0) & 0x0f));
					netentry->fec =             buffer[ptr+descriptorptr+18] & 0x0f;
					
					netsearch = g_slist_prepend((GSList *)netsearch, (gpointer)netentry);
					
					printf("freq: %i, orb. pos: %i, sr: %i, pol: %i, mod: %i, fec: %i\n", netentry->frequency, netentry->position, netentry->sr, netentry->polarization, netentry->modulation, netentry->fec);
				} else {
					printf("5. Descriptorlength: %i, tag 0x%x, length %i\n", ES_info_length, descriptor_tag, descriptor_length);
				}
				
				descriptorptr += descriptor_length;
			}
			
		}
		ptr += transport_descriptors_length + 6;	
	}
	
	netsearch = g_slist_reverse((GSList *)netsearch);
	
	
	//TDT SECTION
	if (status_cb != NULL) status_cb(DVBSCAN_STATUS_TDT);
	
	tdtfilter = start_filter(0x14, 0x70, adapter, demux);
			
	if (read_filter(tdtfilter, (char *)&buffer, sizeof(buffer), DVBSCAN_TDT_READ_TIMEOUT) <= 0) {
		printf("Failed to read TDT table\n");
		return 4;
	}
				
	stop_filter(tdtfilter);
	
	printf("TDT section\n");
		
	table_id = buffer[0];
	table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
	MJD_time = (((buffer[3] & 0xff) << 8) | buffer[4]);
	UTC_time = (((buffer[5] & 0xff) << 16) | ((buffer[6] & 0xff) << 8) | buffer[7]);
		
	y = (guint)((MJD_time - 15078.2) / 365.25);
	m = (guint)((MJD_time - 14956.1 - (guint)(y * 365.25)) / 30.6001);
	d = (guint)(MJD_time - 14956 - (guint)(y * 365.25) - (guint)(m * 30.6001));
	if ((m == 14) || (m == 15)) k = 1; else k = 0;
	y = y + k;
	m = m - 1 - (k * 12);
	
	timestamp = (struct tm *)g_malloc0(sizeof(struct tm));
	timestamp->tm_mday = d;
	timestamp->tm_mon =  m-1;
	timestamp->tm_year = y;
	timestamp->tm_hour = (((UTC_time >> 20) & 0x0F) * 10) + ((UTC_time >> 16) & 0x0F);
	timestamp->tm_min =  (((UTC_time >> 12) & 0x0F) * 10) + ((UTC_time >> 8) & 0x0F);
	timestamp->tm_sec =  (((UTC_time >> 4) & 0x0F) * 10) + (UTC_time & 0x0F);
	
	if (strftime(strbuf, sizeof(strbuf), "%c", timestamp) == 0) {
		printf(" UTC time: broken\n");
	} else {
		printf(" UTC time: %s\n", strbuf);
	}
		
		
	return 5;
}
