/*
 *      alsasound.c
 *      
 *      Copyright 2010 alex <alex@laptop.org>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#define ALSA_PCM_NEW_HW_PARAMS_API
#define ALSA_PCM_NEW_SW_PARAMS_API

#include <glib.h>
#include <gmodule.h>
#include <alsa/asoundlib.h>
#include <sys/time.h>
#include "soundmodules.h"


#define SOUND_MODULE_ID       1
#define SOUND_MODULE_NAME     "ALSA"
#define SOUND_MODULE_VERSION  0x100

#define	ALSA_BUFFER_LEN	      2048
#define	ALSA_DEVICE	          "default"

gboolean deviceopened = FALSE;
snd_pcm_t *device = NULL;


G_MODULE_EXPORT gboolean sound_module_play_file(gchar *filename, gint (*wav_header_parser_callback)(gint filefd, struct _wavheader *wavhdr))
{	
	static gshort buffer[ALSA_BUFFER_LEN];
	gint filefd;
	gint readcount = 0;
	gint bufsize = 0;
	gint multiplex = sizeof(short);
	struct _wavheader wavhdr;
	
	if (wav_header_parser_callback == NULL) return 0;
	
	if ((filefd = open(filename, O_RDONLY, 0)) == -1) {
		printf("Failed to open sound file\n");
		return FALSE;
	}
	
	if (wav_header_parser_callback(filefd, &wavhdr)) {
		if (alsa_open(&wavhdr)) {
			//Correction for 8 bps format
			switch(wavhdr.bits) {
				case 8:  multiplex = sizeof(char); break;
				case 16: multiplex = sizeof(short); break;
				default: multiplex = sizeof(short); break;
			}
			
			//Play sound
			while (wavhdr.len > 0) {
				if (wavhdr.len >= ALSA_BUFFER_LEN*multiplex) {
					bufsize = ALSA_BUFFER_LEN*multiplex;
				} else {
					bufsize = wavhdr.len;
				}
				memset(buffer, 0, sizeof(buffer));
				readcount = read(filefd, &buffer, bufsize);
				wavhdr.len -= readcount;
				alsa_write_data(device, buffer, readcount / ((wavhdr.channels)*multiplex), wavhdr.channels);
			}
		}
	}
	
	close(filefd);
		
	return TRUE;
}

gboolean alsa_open(struct _wavheader *wavhdr)
{	
	const char *devname = "default";
	snd_pcm_hw_params_t *hw_params;
	snd_pcm_sw_params_t *sw_params;
	snd_pcm_uframes_t period_size, buffer_frames;

	int err;
	
	if (!deviceopened) {
		if ((err = snd_pcm_open(&device, devname, SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
			g_warning("ALSA audio device %s open error %s\n", devname, snd_strerror(err));
			if (device != NULL) snd_pcm_close(device);
			deviceopened = FALSE;
			return FALSE;
		}
		deviceopened = TRUE;
	}

	snd_pcm_nonblock(device, 0);

	if ((err = snd_pcm_hw_params_malloc(&hw_params)) < 0) {
		g_warning("ALSA hardware parameters allocation error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_any(device, hw_params)) < 0) {
		g_warning("ALSA hardware parameters initialization error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_set_access(device, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
		g_warning("ALSA access type setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	int format = 0;
	switch(wavhdr->bits) {
		case 8:  format = SND_PCM_FORMAT_U8; break;
		case 16: format = SND_PCM_FORMAT_S16_LE; break;
		default: format = SND_PCM_FORMAT_S16_LE; break;
	}

	if ((err = snd_pcm_hw_params_set_format(device, hw_params, format)) < 0) {
		g_warning("ALSA sample format setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_set_rate_near(device, hw_params, (unsigned int *)&wavhdr->samples_per_sec, 0)) < 0) {
		g_warning("ALSA sample rate setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_set_channels(device, hw_params, wavhdr->channels)) < 0) {
		g_warning("ALSA channel count setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	period_size = 1024;
	buffer_frames = 4*period_size;

	if ((err = snd_pcm_hw_params_set_buffer_size_near(device, hw_params, &buffer_frames)) < 0) {
		g_warning("ALSA buffer size setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_set_period_size_near(device, hw_params, &period_size, 0)) < 0) {
		g_warning("ALSA period size setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params(device, hw_params)) < 0) {
		g_warning("ALSA hardware parameters setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	snd_pcm_hw_params_free(hw_params) ;

	if ((err = snd_pcm_sw_params_malloc(&sw_params)) != 0) {
		g_warning("ALSA software parameters allocation error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_sw_params_current(device, sw_params)) != 0) {
		g_warning("ALSA default software parameters setting setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_sw_params(device, sw_params)) != 0) {
		g_warning("ALSA software parameters setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	snd_pcm_sw_params_free(sw_params);

	snd_pcm_reset(device);

	return TRUE;
}

int alsa_write_data(snd_pcm_t *alsa_dev, short *data, int frames, int channels)
{	
	int wrbytes = 0;
	int wrnow;

	while (wrbytes < frames) {
		wrnow = snd_pcm_writei(alsa_dev, data + wrbytes * channels, frames - wrbytes);

		if (wrnow >= 0) {
			wrbytes += wrnow;
			if (wrbytes == frames) return wrbytes;
			continue;
		}

		switch (wrnow) {
			case -EAGAIN: {
				g_warning("ALSA error: EAGAIN\n");
				continue;
				break;
			}

			case -EPIPE: {
				g_warning("ALSA error: EPIPE\n");
				snd_pcm_prepare(alsa_dev) ;
				break ;
			}

			case -EBADFD: {
				g_warning("ALSA error: BAD FD\n");
				return 0;
				break;
			}

			case -ESTRPIPE: {
				g_warning("ALSA error: ESTRPIPE\n");
				return 0;
				break;
			}

			case -EIO: {
				g_warning("ALSA error: EIO\n");
				return 0;
			}

			default: {
				g_warning("ALSA error: %d\n", wrnow);
				return 0;
				break;
			}
		}
	}
	return wrbytes;
}

G_MODULE_EXPORT gboolean sound_module_close(void)
{
	if ((device != NULL) && (deviceopened)) {
		snd_pcm_drain(device);
		snd_pcm_close(device);
	}
	
	return TRUE;
}

G_MODULE_EXPORT gboolean sound_module_init(guint *moduleid, gchar *modulename, gint *moduleversion)
{
	gsize namelen;
	
	if (access("/proc/asound/cards", R_OK) != 0) return FALSE;
		
	if (moduleid != NULL) {
		*moduleid = SOUND_MODULE_ID;
	}
	
	if (modulename != NULL) {
		namelen = strlen(SOUND_MODULE_NAME);
		memset(modulename, 0, 32);
		if (namelen > 31) {
			strncpy(modulename, SOUND_MODULE_NAME, 31);
		} else {
			strncpy(modulename, SOUND_MODULE_NAME, namelen);
		}
	}
	
	if (moduleversion != NULL) {
		*moduleversion = SOUND_MODULE_VERSION;
	}
	
	return TRUE;
}
