/*
 *      osssound.c
 *      
 *      Copyright 2010 alex <alex@laptop.org>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <gmodule.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include "soundmodules.h"

#define SOUND_MODULE_ID       3
#define SOUND_MODULE_NAME     "OpenSound (OSS)"
#define SOUND_MODULE_VERSION  0x100

#define OPENSOUND_DSP_DEVICE  "/dev/dsp"
#define OPENSOUND_BUFFER_LEN  1024

gchar deviceopened = 0;
gint  device;

gboolean opensound_open(struct _wavheader *wavhdr)
{
	int soundprofile = APF_NORMAL;
	int sampleformat;
	int newsample;
	
	if (!deviceopened) {
		if ((device = open(OPENSOUND_DSP_DEVICE, O_WRONLY, 0)) == -1) {
			printf("Failed to open sound device\n");
			return FALSE;
		}
		deviceopened = TRUE;
	}
	
	switch(wavhdr->bits) {
		case 8:  sampleformat = AFMT_U8; break;
		case 16: sampleformat = AFMT_S16_LE; break;
		default: sampleformat = AFMT_S16_LE; break;
	}
	
	newsample = sampleformat;
	if (ioctl(device, SNDCTL_DSP_SETFMT, &newsample) == -1) {
		printf("Failed to set sample format\n");
		close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	if (newsample != sampleformat) {
		printf("Sample format not supported\n");
		close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	ioctl(device, SNDCTL_DSP_PROFILE, &soundprofile);
	
	if (ioctl(device, SNDCTL_DSP_CHANNELS, &wavhdr->channels) == -1) {
		printf("Failed to set channels number\n");
		close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	if (ioctl(device, SNDCTL_DSP_SPEED, &wavhdr->samples_per_sec) == -1) {
		printf("Failed to set sample rate\n");
		close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	return TRUE;	
}

G_MODULE_EXPORT gboolean sound_module_close(void)
{
	if (deviceopened) {
		ioctl(device, SNDCTL_DSP_SYNC, NULL);
		close(device);
		deviceopened = FALSE;
	}
	
	return TRUE;
}

G_MODULE_EXPORT gboolean sound_module_play_file(gchar *filename, gint (*wav_header_parser_callback)(gint filefd, struct _wavheader *wavhdr))
{	
	static guchar buffer[OPENSOUND_BUFFER_LEN];
	gint readcount = 0;
	gint bufsize = 0;
	gint filefd;
	struct _wavheader wavhdr;
	
	if (wav_header_parser_callback == NULL) return FALSE;
	
	if ((filefd = open(filename, O_RDONLY, 0)) == -1) {
		printf("Failed to open sound file\n");
		return FALSE;
	}
	
	if (wav_header_parser_callback(filefd, &wavhdr)) {
		if (opensound_open(&wavhdr)) {
			while (wavhdr.len > 0) {
				if (wavhdr.len >= OPENSOUND_BUFFER_LEN) {
					bufsize = OPENSOUND_BUFFER_LEN;
				} else {
					bufsize = wavhdr.len;
				}
				memset(buffer, 0, sizeof(buffer));
				readcount = read(filefd, &buffer, bufsize);
				wavhdr.len -= readcount;
				if (write(device, buffer, bufsize) != bufsize) {
					printf("Opensound write error\n");
					close(filefd);
					return FALSE;
				}
			}
		}
	}
	
	ioctl(device, SNDCTL_DSP_SYNC, NULL);
	close(filefd);
	return TRUE;
}

G_MODULE_EXPORT gboolean sound_module_init(guint *moduleid, gchar *modulename, gint *moduleversion)
{
	gint fd;
	gint ver = 0;
	gint namelen;
	
	if ((fd = open(OPENSOUND_DSP_DEVICE, O_RDONLY, 0)) == -1) {
		printf("Failed to open sound device\n");
		return FALSE;
	}

	if (ioctl(fd, OSS_GETVERSION, &ver) == -1) {
		close(fd);
		printf("Failed to get opensound version\n");
		return FALSE;
	}
	
	if (ver < 0x030900) {
		close(fd);
		printf("Wrong opensound version\n");
		return FALSE;
	}
	
	close(fd);
	
	if (moduleid != NULL) {
		*moduleid = SOUND_MODULE_ID;
	}
	
	if (modulename != NULL) {
		namelen = strlen(SOUND_MODULE_NAME);
		memset(modulename, 0, 32);
		if (namelen > 31) {
			strncpy(modulename, SOUND_MODULE_NAME, 31);
		} else {
			strncpy(modulename, SOUND_MODULE_NAME, namelen);
		}
	}
	
	if (moduleversion != NULL) {
		*moduleversion = SOUND_MODULE_VERSION;
	}
	
	return TRUE;
}
