/*
 *      pulsesound.c
 *      
 *      Copyright 2010 alex <alex@laptop.org>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib.h>
#include <gmodule.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <pulse/simple.h>
#include <pulse/error.h>
#include <pulse/gccmacro.h>
#include "soundmodules.h"

#define SOUND_MODULE_ID       2
#define SOUND_MODULE_NAME     "PulseAudio"
#define SOUND_MODULE_VERSION  0x100

#define PULSE_BUFFER_LEN      1024

pa_simple *device = NULL;

gboolean pulseaudio_open(struct _wavheader *wavhdr)
{
	pa_sample_spec ss;
	int error;
	
	switch(wavhdr->bits) {
		case 8:  ss.format = PA_SAMPLE_U8; break;
		case 16: ss.format = PA_SAMPLE_S16LE; break;
		default: ss.format = PA_SAMPLE_S16LE; break;
	}
	
	ss.rate = wavhdr->samples_per_sec;
	ss.channels = wavhdr->channels; 
	
	if ((device = pa_simple_new(NULL, "GSF", PA_STREAM_PLAYBACK, NULL, "playback", &ss, NULL, NULL, &error)) == 0) {
        g_warning("PulseAudio audio server open error %s", pa_strerror(error));
        return FALSE;
    }
	
	return TRUE;	
}

G_MODULE_EXPORT gboolean sound_module_play_file(gchar *filename, gint (*wav_header_parser_callback)(gint filefd, struct _wavheader *wavhdr))
{	
	static uint8_t buffer[PULSE_BUFFER_LEN];
	gint filefd;
	gint readcount = 0;
	gint bufsize = 0;
	gint error;
	struct _wavheader wavhdr;
	
	if (wav_header_parser_callback == NULL) return 0;
	
	if ((filefd = open(filename, O_RDONLY, 0)) == -1) {
		printf("Failed to open sound file");
		return FALSE;
	}
	
	if (wav_header_parser_callback(filefd, &wavhdr)) {
		if (pulseaudio_open(&wavhdr)) {
			//Play sound
			while (wavhdr.len > 0) {
				if (wavhdr.len >= PULSE_BUFFER_LEN) {
					bufsize = PULSE_BUFFER_LEN;
				} else {
					bufsize = wavhdr.len;
				}
				memset(buffer, 0, sizeof(buffer));
				readcount = read(filefd, &buffer, bufsize);
				if (readcount > 0) {
					wavhdr.len -= readcount;
					if (pa_simple_write(device, buffer, (size_t)readcount, &error) < 0) {
						g_warning("PulseAudio audio server write error %s", pa_strerror(error));
						close(filefd);
						return FALSE;
					}
				} else {
					break;
				}
			}
			//drain and close connection
			if (pa_simple_drain(device, &error) < 0) {
				g_warning("PulseAudio audio server drain error %s", pa_strerror(error));
			}
			pa_simple_free(device);
			
			return TRUE;
		}
	}
	
	return FALSE;
}

G_MODULE_EXPORT gboolean sound_module_close(void)
{
	return TRUE;
}

G_MODULE_EXPORT gboolean sound_module_init(guint *moduleid, gchar *modulename, gint *moduleversion)
{
	gchar path[PATH_MAX];
	gchar *tmp;
	gboolean pathfound = FALSE;
	gchar buf[16];
	gint uid = -1;
	DIR *d;
	struct dirent *entry;
	struct stat status;
	gint pidfd;
	gboolean pidres = FALSE;
	gsize namelen;
	
	uid = getuid();
	if ((d = opendir("/tmp")) == NULL) {
		g_warning("Directory listing read error");
		return FALSE;
	}
	
	while (entry = readdir(d)) {
		if (strstr(entry->d_name, "pulse") != NULL) {
			g_sprintf(path, "/tmp/%s", entry->d_name);
			if (stat(path, &status) != -1) {
				if (status.st_uid == uid) {
					pathfound = TRUE;
					break;
				}
			}
		}
	}
	
	closedir(d);
	
	if (pathfound) {
		tmp = g_strdup(path);
		g_sprintf(path, "%s/pid", tmp);
		g_free(tmp);
		if ((pidfd = open(path, O_RDWR, 0)) != -1) {
			if (read(pidfd, &buf, 15) != 0) {
				if (kill((pid_t)atoi(buf), 0) == 0) {
					pidres = TRUE;
				}
			}
			close(pidfd);
		}
	}
	
	if (pidres) {
		if (moduleid != NULL) {
			*moduleid = SOUND_MODULE_ID;
		}
		
		if (modulename != NULL) {
			namelen = strlen(SOUND_MODULE_NAME);
			memset(modulename, 0, 32);
			if (namelen > 31) {
				strncpy(modulename, SOUND_MODULE_NAME, 31);
			} else {
				strncpy(modulename, SOUND_MODULE_NAME, namelen);
			}
		}
		
		if (moduleversion != NULL) {
			*moduleversion = SOUND_MODULE_VERSION;
		}
		
		return TRUE;
	}
	
	return FALSE;
}
